@@ -1,5 +1,5 @@
 /*
- * Copyright(c) 2009 Intel Corporation. All rights reserved.
+ * Copyright(c) 2010 Intel Corporation. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU General Public License,
@@ -17,15 +17,37 @@
  * Maintained at www.Open-FCoE.org
  */
 
-#include "fcoeadm.h"
+#include <sys/param.h>
+#include <sys/ioctl.h>
+#include <linux/types.h>
+#include <ctype.h>
+#include <stddef.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <stdarg.h>
+#include <time.h>
+#include <malloc.h>
+#include <pthread.h>
+#include <limits.h>
+#include <scsi/sg.h>
+#include <byteswap.h>
+#include <net/if.h>
+#include <unistd.h>
+
+#include "net_types.h"
+#include "fc_types.h"
+#include "fc_scsi.h"
+#include "hbaapi.h"
+#include "fcoeadm_display.h"
+#include "fcoe_utils.h"
 
 /* #define TEST_HBAAPI_V1 */
 #ifdef TEST_HBAAPI_V1
-  #define HBA_FCP_SCSI_ENTRY	 HBA_FCPSCSIENTRY
-  #define HBA_FCP_TARGET_MAPPING HBA_FCPTARGETMAPPING
+#define HBA_FCP_SCSI_ENTRY	 HBA_FCPSCSIENTRY
+#define HBA_FCP_TARGET_MAPPING HBA_FCPTARGETMAPPING
 #else
-  #define HBA_FCP_SCSI_ENTRY	 HBA_FCPSCSIENTRYV2
-  #define HBA_FCP_TARGET_MAPPING HBA_FCPTARGETMAPPINGV2
+#define HBA_FCP_SCSI_ENTRY	 HBA_FCPSCSIENTRYV2
+#define HBA_FCP_TARGET_MAPPING HBA_FCPTARGETMAPPINGV2
 #endif
 /* #define TEST_REPORT_LUNS */
 /* #define TEST_READ_CAP_V1 */
@@ -43,11 +65,25 @@
 /* Minimum byte size of the received inquiry data */
 #define MIN_INQ_DATA_SIZE       36
 
+#define FCP_TARG_STR "FCP Target"
+
 struct sa_nameval {
 	char        *nv_name;
 	u_int32_t   nv_val;
 };
 
+/*
+ * HBA and port objects are one-to-one since there
+ * is one host created per Ethernet port (vlan).
+ */
+struct hba_name_table {
+	HBA_HANDLE            hba_handle;
+	HBA_ADAPTERATTRIBUTES hba_attrs;
+	HBA_PORTATTRIBUTES    port_attrs;
+	int                   failed;
+	int                   displayed;
+};
+
 struct sa_nameval port_states[] = {
 	{ "Not Present",    HBA_PORTSTATE_UNKNOWN },
 	{ "Online",         HBA_PORTSTATE_ONLINE },
@@ -84,7 +120,7 @@ struct sa_nameval port_speeds[] = {
  */
 static const char *
 sa_enum_decode(char *buf, size_t len,
-		const struct sa_nameval *tp, u_int32_t val)
+	       const struct sa_nameval *tp, u_int32_t val)
 {
 	snprintf(buf, len, "Unknown");
 	for (; tp->nv_name != NULL; tp++) {
@@ -195,52 +231,59 @@ sa_sys_read_u32(const char *dir, const char *file, u_int32_t *vp)
 	return rc;
 }
 
+static int is_fcp_target(HBA_PORTATTRIBUTES *rp_info)
+{
+	char buf[MAX_STR_LEN];
+
+	if (sa_sys_read_line(rp_info->OSDeviceName, "roles", buf, sizeof(buf)))
+		return -EINVAL;
+
+	if (!strncmp(buf, FCP_TARG_STR, strlen(FCP_TARG_STR)))
+		return 0;
+
+	return -EINVAL;
+}
+
 static void show_wwn(unsigned char *pWwn)
 {
 	sa_dump_wwn(pWwn, 8, 0);
 }
 
-static void
-show_hba_info(int hba_index, HBA_ADAPTERATTRIBUTES *hba_info, int flags)
+static void show_hba_info(HBA_ADAPTERATTRIBUTES *hba_info)
 {
-	printf("\n");
-	printf("HBA #%d\n", hba_index);
 	printf("    Description:      %s\n", hba_info->ModelDescription);
 	printf("    Revision:         %s\n", hba_info->HardwareVersion);
 	printf("    Manufacturer:     %s\n", hba_info->Manufacturer);
 	printf("    Serial Number:    %s\n", hba_info->SerialNumber);
 	printf("    Driver:           %s %s\n", hba_info->DriverName,
-						hba_info->DriverVersion);
+	       hba_info->DriverVersion);
 	printf("    Number of Ports:  %d\n", hba_info->NumberOfPorts);
+	printf("\n");
 }
 
-static void
-show_port_info(int hba_index, int lp_index,
-		HBA_ADAPTERATTRIBUTES *hba_info,
-		HBA_PORTATTRIBUTES *lp_info)
+static void show_port_info(HBA_ADAPTERATTRIBUTES *hba_info,
+			   HBA_PORTATTRIBUTES *lp_info)
 {
 	char buf[256];
 	int len = sizeof(buf);
 
-	printf("\n    Port #%d\n", lp_index);
-
 	printf("        Symbolic Name:     %s\n",
-					lp_info->PortSymbolicName);
+	       lp_info->PortSymbolicName);
 
 	printf("        OS Device Name:    %s\n",
-					lp_info->OSDeviceName);
+	       lp_info->OSDeviceName);
 
 	printf("        Node Name:         0x");
-					show_wwn(lp_info->NodeWWN.wwn);
-					printf("    \n");
+	show_wwn(lp_info->NodeWWN.wwn);
+	printf("\n");
 
 	printf("        Port Name:         0x");
-					show_wwn(lp_info->PortWWN.wwn);
-					printf("    \n");
+	show_wwn(lp_info->PortWWN.wwn);
+	printf("\n");
 
 	printf("        FabricName:        0x");
-					show_wwn(lp_info->FabricName.wwn);
-					printf("    \n");
+	show_wwn(lp_info->FabricName.wwn);
+	printf("\n");
 
 	sa_enum_decode(buf, len, port_speeds, lp_info->PortSpeed);
 	printf("        Speed:             %s\n", buf);
@@ -249,39 +292,39 @@ show_port_info(int hba_index, int lp_index,
 	printf("        Supported Speed:   %s\n", buf);
 
 	printf("        MaxFrameSize:      %d\n",
-					lp_info->PortMaxFrameSize);
+	       lp_info->PortMaxFrameSize);
 
 	printf("        FC-ID (Port ID):   0x%06X\n",
-					lp_info->PortFcId);
+	       lp_info->PortFcId);
 
 	sa_enum_decode(buf, sizeof(buf), port_states, lp_info->PortState);
 	printf("        State:             %s\n", buf);
-
+	printf("\n");
 	/* TODO: Display PortSupportedFc4Types and PortActiveFc4Types */
 }
 
-static void
-show_target_info(int hba_index, int lp_index, int rp_index,
-		HBA_ADAPTERATTRIBUTES *hba_info,
-		HBA_PORTATTRIBUTES *rp_info)
+static void show_target_info(const char *symbolic_name,
+			     HBA_ADAPTERATTRIBUTES *hba_info,
+			     HBA_PORTATTRIBUTES *rp_info)
 {
 	char buf[256];
 	u_int32_t tgt_id;
 	int rc;
+	char *ifname;
 
-	printf("Target #%d @ %s\n",
-				rp_index, hba_info->NodeSymbolicName + 5);
+	ifname = get_ifname_from_symbolic_name(symbolic_name);
 
 	rc = sa_sys_read_line(rp_info->OSDeviceName, "roles", buf, sizeof(buf));
+	printf("    Interface:        %s\n", ifname);
 	printf("    Roles:            %s\n", buf);
 
 	printf("    Node Name:        0x");
-				show_wwn(rp_info->NodeWWN.wwn);
-				printf("    \n");
+	show_wwn(rp_info->NodeWWN.wwn);
+	printf("\n");
 
 	printf("    Port Name:        0x");
-				show_wwn(rp_info->PortWWN.wwn);
-				printf("    \n");
+	show_wwn(rp_info->PortWWN.wwn);
+	printf("\n");
 
 	rc = sa_sys_read_u32(rp_info->OSDeviceName, "scsi_target_id", &tgt_id);
 	if (tgt_id != -1)
@@ -290,14 +333,13 @@ show_target_info(int hba_index, int lp_index, int rp_index,
 	printf("    MaxFrameSize:     %d\n", rp_info->PortMaxFrameSize);
 
 	printf("    OS Device Name:   %s\n",
-				strrchr(rp_info->OSDeviceName, '/') + 1);
+	       strrchr(rp_info->OSDeviceName, '/') + 1);
 
 	printf("    FC-ID (Port ID):  0x%06X\n", rp_info->PortFcId);
 
 	sa_enum_decode(buf, sizeof(buf), port_states, rp_info->PortState);
 	printf("    State:            %s\n", buf);
-
-	printf("    \n");
+	printf("\n");
 }
 
 static void
@@ -395,8 +437,8 @@ get_inquiry_data_v2(HBA_HANDLE hba_handle,
 #ifdef TEST_HBAAPI_V1
 static HBA_STATUS
 get_device_capacity_v1(HBA_HANDLE hba_handle,
-		    HBA_FCPSCSIENTRY *ep,
-		    char *buf, size_t len)
+		       HBA_FCPSCSIENTRY *ep,
+		       char *buf, size_t len)
 {
 	char sense[128];
 	HBA_UINT32 rlen;
@@ -433,9 +475,9 @@ get_device_capacity_v1(HBA_HANDLE hba_handle,
 #else
 static HBA_STATUS
 get_device_capacity_v2(HBA_HANDLE hba_handle,
-		    HBA_PORTATTRIBUTES *lp_info,
-		    HBA_FCPSCSIENTRYV2 *ep,
-		    char *buf, size_t len)
+		       HBA_PORTATTRIBUTES *lp_info,
+		       HBA_FCPSCSIENTRYV2 *ep,
+		       char *buf, size_t len)
 {
 	char sense[128];
 	HBA_UINT32 rlen;
@@ -461,7 +503,7 @@ get_device_capacity_v2(HBA_HANDLE hba_handle,
 						&slen);
 		if ((status == HBA_STATUS_OK) && (sstat == SCSI_ST_GOOD))
 			return HBA_STATUS_OK;
-		if ((sstat == SCSI_ST_CHECK) && (sense[2] == 0x06))
+		if (sstat == SCSI_ST_CHECK)
 			continue;
 		fprintf(stderr,
 			"%s: HBA_ScsiReadCapacityV2 failed, "
@@ -512,7 +554,7 @@ get_device_serial_number(HBA_HANDLE hba_handle,
 	unit_sn = (struct scsi_inquiry_unit_sn *)rspbuf;
 	unit_sn->is_serial[unit_sn->is_page_len] = '\0';
 	sa_strncpy_safe(buf, buflen, (char *)unit_sn->is_serial,
-				     (size_t)unit_sn->is_page_len);
+			(size_t)unit_sn->is_page_len);
 	return HBA_STATUS_OK;
 }
 #endif
@@ -555,11 +597,11 @@ get_report_luns_data_v1(HBA_HANDLE hba_handle, HBA_FCPSCSIENTRYV2 *ep)
 		rlen = (HBA_UINT32) sizeof(rspbuf);
 		slen = (HBA_UINT32) sizeof(sense);
 		status = HBA_SendReportLUNs(hba_handle,
-					     ep->FcpId.PortWWN,
-					     rspbuf,
-					     rlen,
-					     sense,
-					     slen);
+					    ep->FcpId.PortWWN,
+					    rspbuf,
+					    rlen,
+					    sense,
+					    slen);
 		if (status == HBA_STATUS_OK) {
 			show_report_luns_data(rspbuf);
 			return HBA_STATUS_OK;
@@ -579,8 +621,8 @@ get_report_luns_data_v1(HBA_HANDLE hba_handle, HBA_FCPSCSIENTRYV2 *ep)
 
 static HBA_STATUS
 get_report_luns_data_v2(HBA_HANDLE hba_handle,
-		       HBA_PORTATTRIBUTES *lp_info,
-		       HBA_FCPSCSIENTRYV2 *ep)
+			HBA_PORTATTRIBUTES *lp_info,
+			HBA_FCPSCSIENTRYV2 *ep)
 {
 	HBA_STATUS status;
 	char rspbuf[512 * sizeof(u_int64_t)]; /* max 512 luns */
@@ -627,9 +669,9 @@ static void
 show_short_lun_info_header(void)
 {
 	printf("    LUN ID  Device Name   Capacity   "
-		"Block Size  Description\n");
+	       "Block Size  Description\n");
 	printf("    ------  -----------  ----------  ----------  "
-		"----------------------------\n");
+	       "----------------------------\n");
 }
 
 static void
@@ -652,7 +694,7 @@ show_short_lun_info(HBA_FCP_SCSI_ENTRY *ep, char *inqbuf,
 
 	/* Get device capacity */
 	cap = (u_int64_t) net32_get(&rcap_resp->rc_block_len) *
-			  net32_get(&rcap_resp->rc_lba);
+		net32_get(&rcap_resp->rc_lba);
 	cap_abbr = cap / (1024.0 * 1024.0);
 	abbr = "MB";
 	if (cap_abbr >= 1024) {
@@ -679,9 +721,9 @@ show_short_lun_info(HBA_FCP_SCSI_ENTRY *ep, char *inqbuf,
 
 	/* Show the LUN info */
 	printf("%10d  %-11s  %10s  %7d     %s %s (rev %s)\n",
-		ep->ScsiId.ScsiOSLun, ep->ScsiId.OSDeviceName,
-		capstr, net32_get(&rcap_resp->rc_block_len),
-		vendor, model, rev);
+	       ep->ScsiId.ScsiOSLun, ep->ScsiId.OSDeviceName,
+	       capstr, net32_get(&rcap_resp->rc_block_len),
+	       vendor, model, rev);
 }
 
 static void
@@ -723,7 +765,7 @@ show_full_lun_info(HBA_HANDLE hba_handle,
 
 	/* Get device capacity */
 	cap = (u_int64_t) net32_get(&rcap_resp->rc_block_len) *
-			  net32_get(&rcap_resp->rc_lba);
+		net32_get(&rcap_resp->rc_lba);
 	cap_abbr = cap / (1024.0 * 1024.0);
 	abbr = "MB";
 	if (cap_abbr >= 1024) {
@@ -747,25 +789,25 @@ show_full_lun_info(HBA_HANDLE hba_handle,
 	/* Show lun info */
 	printf("    LUN #%d Information:\n", ep->ScsiId.ScsiOSLun);
 	printf("        OS Device Name:     %s\n",
-					ep->ScsiId.OSDeviceName);
+	       ep->ScsiId.OSDeviceName);
 	printf("        Description:        %s %s (rev %s)\n",
-					vendor, model, rev);
+	       vendor, model, rev);
 	printf("        Ethernet Port FCID: 0x%06X\n",
-					lp_info->PortFcId);
+	       lp_info->PortFcId);
 	printf("        Target FCID:        0x%06X\n",
-					rp_info->PortFcId);
+	       rp_info->PortFcId);
 	if (tgt_id == -1)
 		printf("        Target ID:          (None)\n");
 	else
 		printf("        Target ID:          %u\n", tgt_id);
 	printf("        LUN ID:             %d\n",
-					ep->ScsiId.ScsiOSLun);
+	       ep->ScsiId.ScsiOSLun);
 
 	printf("        Capacity:           %s\n", capstr);
 	printf("        Capacity in Blocks: %d\n",
-					net32_get(&rcap_resp->rc_lba));
+	       net32_get(&rcap_resp->rc_lba));
 	printf("        Block Size:         %d bytes\n",
-					net32_get(&rcap_resp->rc_block_len));
+	       net32_get(&rcap_resp->rc_block_len));
 	pqual = inq->is_periph & SCSI_INQ_PQUAL_MASK;
 	if (pqual == SCSI_PQUAL_ATT)
 		printf("        Status:             Attached\n");
@@ -773,7 +815,7 @@ show_full_lun_info(HBA_HANDLE hba_handle,
 		printf("        Status:             Detached\n");
 	else if (pqual == SCSI_PQUAL_NC)
 		printf("        Status:             "
-			"Not capable of attachment\n");
+		       "Not capable of attachment\n");
 
 #ifdef TEST_DEV_SERIAL_NO
 	/* Show the serial number of the device */
@@ -833,7 +875,7 @@ get_device_map(HBA_HANDLE hba_handle, HBA_PORTATTRIBUTES *lp_info,
 		status = HBA_GetFcpTargetMapping(hba_handle, map);
 #else
 		status = HBA_GetFcpTargetMappingV2(
-				hba_handle, lp_info->PortWWN, map);
+			hba_handle, lp_info->PortWWN, map);
 #endif
 		if (map->NumberOfEntries > limit) {
 			limit = map->NumberOfEntries + LUN_COUNT_INCR;
@@ -877,7 +919,8 @@ scan_device_map(HBA_HANDLE hba_handle,
 		HBA_ADAPTERATTRIBUTES *hba_info,
 		HBA_PORTATTRIBUTES *lp_info,
 		HBA_PORTATTRIBUTES *rp_info,
-		struct opt_info *opt_info)
+		const char *ifname,
+		enum disp_style style)
 {
 	HBA_STATUS status;
 	HBA_FCP_TARGET_MAPPING *map = NULL;
@@ -901,12 +944,6 @@ scan_device_map(HBA_HANDLE hba_handle,
 		if (ep->FcpId.FcId != rp_info->PortFcId)
 			continue;
 
-		if (opt_info->l_flag &&
-		    opt_info->l_fcid_present &&
-		    opt_info->l_lun_id_present &&
-		    ep->ScsiId.ScsiOSLun != opt_info->l_lun_id)
-			continue;
-
 		dev = ep->ScsiId.OSDeviceName;
 		if (strstr(dev, "/dev/") == dev)
 			dev += 5;
@@ -914,10 +951,10 @@ scan_device_map(HBA_HANDLE hba_handle,
 		/* Issue standard inquiry */
 #ifdef TEST_HBAAPI_V1
 		status = get_inquiry_data_v1(hba_handle, ep,
-					inqbuf, sizeof(inqbuf));
+					     inqbuf, sizeof(inqbuf));
 #else
 		status = get_inquiry_data_v2(hba_handle, lp_info,
-					ep, inqbuf, sizeof(inqbuf));
+					     ep, inqbuf, sizeof(inqbuf));
 #endif
 		if (status != HBA_STATUS_OK)
 			continue;
@@ -926,53 +963,59 @@ scan_device_map(HBA_HANDLE hba_handle,
 		/* Issue read capacity */
 #ifdef TEST_HBAAPI_V1
 		status = get_device_capacity_v1(hba_handle, ep,
-					(char *)&rcap_resp, sizeof(rcap_resp));
+						(char *)&rcap_resp,
+						sizeof(rcap_resp));
 #else
 		status = get_device_capacity_v2(hba_handle, lp_info,
-					ep, (char *)&rcap_resp,
-					sizeof(rcap_resp));
+						ep, (char *)&rcap_resp,
+						sizeof(rcap_resp));
 #endif
 		if (status != HBA_STATUS_OK)
 			continue;
-
-		if (opt_info->t_flag) {
+		switch (style) {
+		case DISP_TARG:
 			if (!print_header) {
 				show_short_lun_info_header();
 				print_header = 1;
 			}
 			show_short_lun_info(ep, inqbuf, &rcap_resp);
-		} else if (opt_info->l_flag)
+			break;
+		case DISP_LUN:
 			show_full_lun_info(hba_handle, hba_info, lp_info,
-				rp_info, ep, inqbuf, &rcap_resp);
+					   rp_info, ep, inqbuf, &rcap_resp);
+			break;
+		}
 
 #ifdef TEST_REPORT_LUNS
 		if (i == 0) {	/* only issue report luns to the first LUN */
- #ifdef TEST_HBAAPI_V1
+#ifdef TEST_HBAAPI_V1
 			get_report_luns_data_v1(hba_handle, ep);
- #else
+#else
 			get_report_luns_data_v2(hba_handle, lp_info, ep);
- #endif
+#endif
 		}
 #endif
 	}
 
+	/* Newline at the end of the short lun report */
+	if (style == DISP_TARG)
+		printf("\n");
+
 	free(map);
-	printf("\n");
 }
 
-static void
-show_port_stats_header(struct opt_info *opt_info)
+static void show_port_stats_header(const char *ifname, int interval)
 {
 	printf("\n");
 	printf("%-7s interval: %-2d                                    Err  Inv  "
-		"IvTx Link Cntl Input     Input     Output    Output\n",
-		 opt_info->ifname, opt_info->n_interval);
+	       "IvTx Link Cntl Input     Input     Output    Output\n",
+	       ifname, interval);
 	printf("Seconds TxFrames  TxBytes      RxFrames  RxBytes        "
-		"Frms CRC  Byte Fail Reqs Requests  MBytes    "
-		"Requests  MBytes\n");
+	       "Frms CRC  Byte Fail Reqs Requests  MBytes    "
+	       "Requests  MBytes\n");
 	printf("------- --------- ------------ --------- -------------- "
-		"---- ---- ---- ---- ---- --------- --------- "
-		"--------- ---------\n");
+	       "---- ---- ---- ---- ---- --------- --------- "
+	       "--------- ---------\n");
 }
 
 static void
@@ -997,44 +1040,74 @@ show_port_stats_in_row(HBA_INT64 start_time,
 	printf("\n");
 }
 
+static void hba_table_destroy(struct hba_name_table *hba_table)
+{
+	int i;
 
-void
-display_port_stats(struct opt_info *opt_info)
+	/*
+	 * This is inefficiant as is closes adapter handles
+	 * (presumedly 0) that weren't open. However, it allows
+	 * us to avoid maintaining a count.
+	 */
+	for (i = 0 ; i < MAX_HBA_COUNT ; i++)
+		HBA_CloseAdapter(hba_table[i].hba_handle);
+}
+
+static enum fcoe_status fcoeadm_loadhba()
+{
+	if (HBA_STATUS_OK != HBA_LoadLibrary())
+		return EHBAAPIERR;
+
+	return SUCCESS;
+}
+
+/*
+ * This routine leaves all adapters fd's open.
+ */
+static int hba_table_init(struct hba_name_table *hba_table)
 {
 	HBA_STATUS retval;
-	HBA_UINT32 hba_cnt;
-	HBA_HANDLE hba_handle;
-	HBA_ADAPTERATTRIBUTES hba_attrs;
-	HBA_PORTATTRIBUTES port_attrs;
-	HBA_PORTSTATISTICS port_stats;
-	HBA_FC4STATISTICS port_fc4stats;
-	HBA_INT64 start_time = 0;
-	char namebuf[1028];
-	int i = 0, found = 0;
+	char namebuf[1024];
+	int i, num_hbas = 0;
 
-	hba_cnt = HBA_GetNumberOfAdapters();
-	if (!hba_cnt) {
-		fprintf(stderr, "No FCoE interfaces created.\n");
-		return;
+	/*
+	 * Initialize the table.
+	 */
+	for (i = 0 ; i < MAX_HBA_COUNT ; i++) {
+		memset(&hba_table[i], 0,
+		       sizeof(struct hba_name_table));
 	}
 
-	for (i = 0; i < hba_cnt; i++) {
+	num_hbas = HBA_GetNumberOfAdapters();
+	if (!num_hbas) {
+		fprintf(stderr, "No FCoE interfaces created.\n");
+		return num_hbas;
+	}
+
+	/*
+	 * Fill out the HBA table.
+	 */
+	for (i = 0; i < num_hbas ; i++) {
 		retval = HBA_GetAdapterName(i, namebuf);
 		if (retval != HBA_STATUS_OK) {
-			fprintf(stderr, "Failure of HBA_GetAdapterName: %d\n",
-				retval);
+			fprintf(stderr,
+				"Failure of HBA_GetAdapterName: %d\n", retval);
 			continue;
 		}
 
-		hba_handle = HBA_OpenAdapter(namebuf);
-		if (!hba_handle) {
+		hba_table[i].hba_handle = HBA_OpenAdapter(namebuf);
+		if (!hba_table[i].hba_handle) {
+			hba_table[i].failed = 1;
 			fprintf(stderr, "HBA_OpenAdapter failed\n");
 			perror("HBA_OpenAdapter");
 			continue;
 		}
 
-		retval = HBA_GetAdapterAttributes(hba_handle, &hba_attrs);
+		retval = HBA_GetAdapterAttributes(hba_table[i].hba_handle,
+						  &hba_table[i].hba_attrs);
 		if (retval != HBA_STATUS_OK) {
+			HBA_CloseAdapter(hba_table[i].hba_handle);
+			hba_table[i].failed = 1;
 			fprintf(stderr,
 				"HBA_GetAdapterAttributes failed, retval=%d\n",
 				retval);
@@ -1042,32 +1115,77 @@ display_port_stats(struct opt_info *opt_info)
 			continue;
 		}
 
-		retval = HBA_GetAdapterPortAttributes(
-					hba_handle, 0, &port_attrs);
+		retval = HBA_GetAdapterPortAttributes(hba_table[i].hba_handle,
+						      0,
+						      &hba_table[i].port_attrs);
 		if (retval != HBA_STATUS_OK) {
+			HBA_CloseAdapter(hba_table[i].hba_handle);
+			hba_table[i].failed = 1;
 			fprintf(stderr,
 				"HBA_GetAdapterPortAttributes failed, "
-				"status=%d\n", retval);
+				"retval=%d\n", retval);
 			continue;
 		}
-
-		if (strstr(port_attrs.PortSymbolicName, opt_info->ifname)) {
-			found = 1;
-			break;
-		}
 	}
 
-	if (!found) {
-		fprintf(stderr, "Cannot find attributes for %s\n",
-			opt_info->ifname);
-		return;
+	return num_hbas;
+}
+
+/*
+ * This routine expects a valid interface name.
+ */
+static int get_index_for_ifname(struct hba_name_table *hba_table,
+				int num_hbas, const char *ifname)
+{
+	int i;
+
+	for (i = 0 ; i < num_hbas ; i++) {
+		if (!check_symbolic_name_for_interface(
+			    hba_table[i].port_attrs.PortSymbolicName,
+			    ifname))
+			return i;
 	}
 
+	return -EINVAL;
+}
+
+enum fcoe_status display_port_stats(const char *ifname, int interval)
+{
+	HBA_STATUS retval;
+	HBA_HANDLE hba_handle;
+	HBA_PORTATTRIBUTES *port_attrs;
+	HBA_PORTSTATISTICS port_stats;
+	HBA_FC4STATISTICS port_fc4stats;
+	HBA_INT64 start_time = 0;
+	struct hba_name_table hba_table[MAX_HBA_COUNT];
+	enum fcoe_status rc = SUCCESS;
+	int i, num_hbas;
+
+	if (fcoeadm_loadhba())
+		return EHBAAPIERR;
+
+	num_hbas = hba_table_init(hba_table);
+
+	i = get_index_for_ifname(hba_table, num_hbas, ifname);
+
+	/*
+	 * Return error code if a valid index wasn't returned.
+	 */
+	if (i < 0) {
+		hba_table_destroy(hba_table);
+		HBA_FreeLibrary();
+		return EHBAAPIERR;
+	}
+
+	hba_handle = hba_table[i].hba_handle;
+	port_attrs = &hba_table[i].port_attrs;
+
 	i = 0;
 	while (1) {
 		unsigned int secs_left;
 
-		retval = HBA_GetPortStatistics(hba_handle, 0, &port_stats);
+		retval = HBA_GetPortStatistics(hba_handle,
+					       0, &port_stats);
 		if (retval != HBA_STATUS_OK &&
 		    retval != HBA_STATUS_ERROR_NOT_SUPPORTED) {
 			fprintf(stderr,
@@ -1078,7 +1196,7 @@ display_port_stats(struct opt_info *opt_info)
 		if (retval == HBA_STATUS_ERROR_NOT_SUPPORTED) {
 			fprintf(stderr,
 				"Port Statistics not supported by %s\n",
-				opt_info->ifname);
+				ifname);
 			break;
 		}
 
@@ -1086,258 +1204,188 @@ display_port_stats(struct opt_info *opt_info)
 			start_time = port_stats.SecondsSinceLastReset;
 
 		retval = HBA_GetFC4Statistics(hba_handle,
-					      port_attrs.PortWWN,
+					      port_attrs->PortWWN,
 					      FC_TYPE_FCP,
 					      &port_fc4stats);
 		if (retval != HBA_STATUS_OK &&
 		    retval != HBA_STATUS_ERROR_NOT_SUPPORTED) {
 			fprintf(stderr, "HBA_GetFC4Statistics failed, "
-				 "status=%d\n", retval);
+				"status=%d\n", retval);
 			break;
 		}
 		if (retval == HBA_STATUS_ERROR_NOT_SUPPORTED) {
 			fprintf(stderr,
 				"Port FC4 Statistics not supported by %s\n",
-				opt_info->ifname);
+				ifname);
 			break;
 		}
 		if (!(i % 52))
-			show_port_stats_header(opt_info);
+			show_port_stats_header(ifname, interval);
 		show_port_stats_in_row(start_time, &port_stats, &port_fc4stats);
 		i++;
 
 		/* wait for the requested time interval in seconds */
-		secs_left = opt_info->n_interval;
+		secs_left = interval;
 		do {
 			secs_left = sleep(secs_left);
 		} while (secs_left);
 	}
 
-	HBA_CloseAdapter(hba_handle);
-	return;
+	hba_table_destroy(hba_table);
+	HBA_FreeLibrary();
+	return rc;
 }
 
-static struct hba_name_table {
-	char  SerialNumber[64];
-	int   index;
-} hba_name_table[MAX_HBA_COUNT];
-
-static int
-find_hba_index(char *serial_number, int *hba_index)
+enum fcoe_status display_adapter_info(const char *ifname)
 {
-	int i, j;
+	struct hba_name_table hba_table[MAX_HBA_COUNT];
+	enum fcoe_status rc = SUCCESS;
+	int i, j, num_hbas = 0;
 
-	j = sizeof(hba_name_table[0].SerialNumber) - 1;
-	for (i = 0; i < MAX_HBA_COUNT; i++) {
-		if (hba_name_table[i].index == -1) {
-			/* not found */
-			hba_name_table[i].index = i;
-			/* TODO: change to sa_strncpy_safe */
-			strncpy(hba_name_table[i].SerialNumber,
-				serial_number, j);
-			*hba_index = i;
-			return 1;    /* print hba info */
-		}
-		if (!strncmp(serial_number,
-		    hba_name_table[i].SerialNumber, j)) {
-			*hba_index = hba_name_table[i].index;
-			return 0;    /* do not print hba info */
-		}
-	}
-	/* table full */
-	return -1;
-}
+	if (fcoeadm_loadhba())
+		return EHBAAPIERR;
 
-void
-display_adapter_info(struct opt_info *opt_info)
-{
-	HBA_STATUS retval;
-	HBA_UINT32 hba_cnt;
-	HBA_UINT32 lport_cnt_per_hba = 1;  /* always one port per hba */
-	HBA_HANDLE hba_handle;
-	HBA_ADAPTERATTRIBUTES hba_attrs;
-	HBA_PORTATTRIBUTES port_attrs;
-	char namebuf[1028];
-	int i, j, rc;
-	int hba_index = -1;
-	int lp_index = -1;
+	num_hbas = hba_table_init(hba_table);
 
-	for (i = 0; i < MAX_HBA_COUNT; i++) {
-		hba_name_table[i].index = -1;
-		memset(hba_name_table[i].SerialNumber, 0,
-			sizeof(hba_name_table[i].SerialNumber));
-	}
+	/*
+	 * Loop through each HBA entry and for each serial number
+	 * not already printed print the header and each sub-port
+	 * on that adapter.
+	 */
+	for (i = 0 ; i < num_hbas ; i++) {
+		if (hba_table[i].failed ||
+		    hba_table[i].displayed)
+			continue;
 
-	hba_cnt = HBA_GetNumberOfAdapters();
-	if (!hba_cnt) {
-		fprintf(stderr, "No FCoE interfaces created.\n");
-		return;
-	}
-
-	for (i = 0; i < hba_cnt; i++) {
-		retval = HBA_GetAdapterName(i, namebuf);
-		if (retval != HBA_STATUS_OK) {
-			fprintf(stderr,
-				"Failure of HBA_GetAdapterName: %d\n", retval);
+		if (ifname && check_symbolic_name_for_interface(
+			    hba_table[i].port_attrs.PortSymbolicName,
+			    ifname)) {
+			/*
+			 * Overloading 'displayed' to indicate
+			 * that the HBA/Port should be skipped.
+			 */
+			hba_table[i].displayed = 1;
 			continue;
 		}
 
-		hba_handle = HBA_OpenAdapter(namebuf);
-		if (!hba_handle) {
-			fprintf(stderr, "HBA_OpenAdapter failed\n");
-			perror("HBA_OpenAdapter");
-			continue;
-		}
-
-		retval = HBA_GetAdapterAttributes(hba_handle, &hba_attrs);
-		if (retval != HBA_STATUS_OK) {
-			fprintf(stderr,
-				"HBA_GetAdapterAttributes failed, retval=%d\n",
-				retval);
-			perror("HBA_GetAdapterAttributes");
-			continue;
-		}
-
-		rc = find_hba_index(hba_attrs.SerialNumber, &hba_index);
-		if (rc == -1) {
-			fprintf(stderr,
-				"Too many adapters. Maximum %d\n",
-				MAX_HBA_COUNT);
-			return;
-		} else if (rc == 1)
-			show_hba_info(hba_index, &hba_attrs, 0);
-
-		for (j = 0; j < lport_cnt_per_hba; j++) {
-			retval = HBA_GetAdapterPortAttributes(
-					hba_handle, j, &port_attrs);
-			if (retval != HBA_STATUS_OK) {
-				fprintf(stderr,
-					"HBA_GetAdapterPortAttributes failed, "
-					"j=%d, status=%d\n", j, retval);
-				continue;
-			}
-
-			lp_index++;
-			if (opt_info->ifname &&
-			    !strstr(port_attrs.PortSymbolicName,
-					opt_info->ifname))
-				continue;
-			show_port_info(hba_index, lp_index, &hba_attrs,
-					&port_attrs);
-		}
-		HBA_CloseAdapter(hba_handle);
-	}
-}
-
-void
-display_target_info(struct opt_info *opt_info)
-{
-	HBA_STATUS retval;
-	HBA_UINT32 hba_cnt;
-	HBA_UINT32 lport_cnt_per_hba = 1;  /* always one port per hba */
-	HBA_HANDLE hba_handle;
-	HBA_ADAPTERATTRIBUTES hba_attrs;
-	HBA_PORTATTRIBUTES port_attrs;
-	HBA_PORTATTRIBUTES rport_attrs;
-	char namebuf[1028];
-	int i, j, rc;
-	int hba_index = -1;
-	int lp_index = -1;
-	int rp_index = -1;
-
-	for (i = 0; i < MAX_HBA_COUNT; i++) {
-		hba_name_table[i].index = -1;
-		memset(hba_name_table[i].SerialNumber, 0,
-			sizeof(hba_name_table[i].SerialNumber));
-	}
-
-	hba_cnt = HBA_GetNumberOfAdapters();
-	if (!hba_cnt) {
-		fprintf(stderr, "No FCoE interfaces created.\n");
-		return;
-	}
-
-	for (i = 0; i < hba_cnt; i++) {
-		retval = HBA_GetAdapterName(i, namebuf);
-		if (retval != HBA_STATUS_OK) {
-			fprintf(stderr,
-				"Failure of HBA_GetAdapterName: %d\n", retval);
-			continue;
-		}
-
-		hba_handle = HBA_OpenAdapter(namebuf);
-		if (!hba_handle) {
-			fprintf(stderr, "HBA_OpenAdapter failed\n");
-			perror("HBA_OpenAdapter");
-			continue;
-		}
-
-		retval = HBA_GetAdapterAttributes(hba_handle, &hba_attrs);
-		if (retval != HBA_STATUS_OK) {
-			fprintf(stderr,
-				"HBA_GetAdapterAttributes failed, retval=%d\n",
-				retval);
-			perror("HBA_GetAdapterAttributes");
-			continue;
-		}
-
-		rc = find_hba_index(hba_attrs.SerialNumber, &hba_index);
-		if (rc == -1) {
-			fprintf(stderr,
-				"Too many adapters. Maximum %d\n",
-				MAX_HBA_COUNT);
-			return;
-		}
-
-		for (j = 0; j < lport_cnt_per_hba; j++) {
-			retval = HBA_GetAdapterPortAttributes(
-					hba_handle, j, &port_attrs);
-			if (retval != HBA_STATUS_OK) {
-				fprintf(stderr,
-					"HBA_GetAdapterPortAttributes failed, "
-					"j=%d, status=%d\n", j, retval);
-				continue;
-			}
-
-			lp_index++;
-			if (opt_info->ifname &&
-			    !strstr(port_attrs.PortSymbolicName,
-					opt_info->ifname))
-				continue;
-
-			for (rp_index = 0;
-			     rp_index < port_attrs.NumberofDiscoveredPorts;
-			     rp_index++) {
-				retval = HBA_GetDiscoveredPortAttributes(
-						hba_handle, j, rp_index,
-						&rport_attrs);
-				if (retval != HBA_STATUS_OK) {
-					fprintf(stderr,
-					"HBA_GetDiscoveredPortAttributes "
-					"failed, j=%d, for rp_index=%d, "
-					"status=%d\n", j, rp_index, retval);
-					continue;
-				}
+		/*
+		 * Display the adapter header.
+		 */
+		show_hba_info(&hba_table[i].hba_attrs);
 
+		/*
+		 * Loop through HBAs again to print sub-ports.
+		 */
+		for (j = 0; j < num_hbas ; j++) {
+			if (ifname && check_symbolic_name_for_interface(
+				    hba_table[j].port_attrs.PortSymbolicName,
+				    ifname)) {
 				/*
-				 * If -l option and fcid are specified in the
-				 * command, filter out the targets do not have
-				 * port ID equals to fcid.
+				 * Overloading 'displayed' to indicate
+				 * that the HBA/Port should be skipped.
 				 */
-				if (opt_info->l_flag &&
-				    opt_info->l_fcid_present &&
-				    rport_attrs.PortFcId != opt_info->l_fcid)
-					continue;
+				hba_table[i].displayed = 1;
+				continue;
+			}
 
-				show_target_info(hba_index, lp_index,
-						 rp_index, &hba_attrs,
-						 &rport_attrs);
-
-				scan_device_map(hba_handle, &hba_attrs,
-						&port_attrs, &rport_attrs,
-						opt_info);
+			if (!strncmp(hba_table[i].hba_attrs.SerialNumber,
+				     hba_table[j].hba_attrs.SerialNumber,
+				     strlen(hba_table[i].hba_attrs.SerialNumber))) {
+				show_port_info(&hba_table[j].hba_attrs,
+					       &hba_table[j].port_attrs);
+				hba_table[j].displayed = 1;
 			}
 		}
-		HBA_CloseAdapter(hba_handle);
 	}
+
+	hba_table_destroy(hba_table);
+	HBA_FreeLibrary();
+
+	return rc;
+}
+
+enum fcoe_status display_target_info(const char *ifname,
+				     enum disp_style style)
+{
+	HBA_STATUS retval;
+	HBA_PORTATTRIBUTES rport_attrs;
+	struct hba_name_table hba_table[MAX_HBA_COUNT];
+	int i, target_index, num_hbas = 0;
+	enum fcoe_status rc = SUCCESS;
+
+	if (fcoeadm_loadhba())
+		return EHBAAPIERR;
+
+	num_hbas = hba_table_init(hba_table);
+
+	/*
+	 * Loop through each HBA entry and for each serial number
+	 * not already printed print the header and each sub-port
+	 * on that adapter.
+	 */
+	for (i = 0 ; i < num_hbas ; i++) {
+		if (hba_table[i].failed ||
+		    hba_table[i].displayed)
+			continue;
+
+		if (ifname && check_symbolic_name_for_interface(
+			    hba_table[i].port_attrs.PortSymbolicName,
+			    ifname)) {
+			/*
+			 * Overloading 'displayed' to indicate
+			 * that the HBA/Port should be skipped.
+			 */
+			hba_table[i].displayed = 1;
+			continue;
+		}
+
+		for (target_index = 0;
+		     target_index < hba_table[i].port_attrs.NumberofDiscoveredPorts;
+		     target_index++) {
+
+			/* TODO: Second arg might be incorrect */
+			retval = HBA_GetDiscoveredPortAttributes(
+				hba_table[i].hba_handle,
+				0, target_index,
+				&rport_attrs);
+
+			if (retval != HBA_STATUS_OK) {
+				fprintf(stderr,
+					"HBA_GetDiscoveredPortAttributes "
+					"failed for target_index=%d, "
+					"status=%d\n", target_index, retval);
+				hba_table[i].failed = 1;
+				continue;
+			}
+
+			/*
+			 * Skip any targets that are not FCP targets
+			 */
+			if (is_fcp_target(&rport_attrs))
+				continue;
+
+			show_target_info(
+				hba_table[i].port_attrs.PortSymbolicName,
+				&hba_table[i].hba_attrs,
+				&rport_attrs);
+
+			if (hba_table[i].port_attrs.PortState !=
+			    HBA_PORTSTATE_ONLINE)
+				continue;
+
+			/*
+			 * This will print the LUN table
+			 * under the target.
+			 */
+			scan_device_map(hba_table[i].hba_handle,
+					&hba_table[i].hba_attrs,
+					&hba_table[i].port_attrs,
+					&rport_attrs, ifname, style);
+		}
+	}
+
+	hba_table_destroy(hba_table);
+	HBA_FreeLibrary();
+
+	return rc;
 }
